import torch

import pytest
import torchlayers


class ConcatenateProxy(torch.nn.Module):
    # Return same tensor three times
    # You could explicitly return a list or tuple as well
    def forward(self, tensor):
        return tensor, tensor, tensor


class _CustomLinearImpl(torch.nn.Linear):
    def __init__(self, in_features, out_features, bias: bool = True):
        super().__init__(in_features, out_features, bias)
        self.some_params = torch.nn.Parameter(torch.randn(2, out_features))


CustomLinear = torchlayers.infer(_CustomLinearImpl)


@pytest.fixture
def model():
    return torchlayers.Sequential(
        torchlayers.Conv(64),
        torchlayers.ReLU(),
        torchlayers.MaxPool(),
        torchlayers.Conv(128),
        torchlayers.ReLU(),
        torchlayers.MaxPool(),
        torchlayers.Conv(256),
        torchlayers.ReLU(),
        torchlayers.Reshape(-1),
    )


def test_reshape(model):
    assert model(torch.randn(16, 1, 32, 32)).shape == (16, 256 * 8 * 8)


def test_lambda():
    layer = torchlayers.Lambda(lambda inputs: inputs * 3)
    output = layer(torch.ones(16))
    assert torch.sum(output) == 16 * 3


def test_concatenate():
    model = torch.nn.Sequential(ConcatenateProxy(), torchlayers.Concatenate(dim=-1))
    assert model(torch.randn(64, 20)).shape == torch.randn(64, 60).shape


def test_custom_inferable_output():
    layer = CustomLinear(out_features=32)
    assert layer(torch.rand(16, 64)).shape == torch.randn(16, 32).shape


def test_custom_inferable_parameters():
    layer = CustomLinear(32)
    layer(torch.rand(16, 64))
    assert layer.some_params.shape == (2, 32)


def test_custom_inferable_build():
    layer = CustomLinear(32)
    layer = torchlayers.build(layer, torch.rand(16, 64))
    assert layer.some_params.shape == (2, 32)



import torch

import pytest
import torchlayers


def test_gru():
    model = torchlayers.build(
        torchlayers.GRU(hidden_size=40, dropout=0.5, batch_first=True, num_layers=3),
        torch.randn(5, 3, 10),
    )

    output, _ = model(torch.randn(5, 3, 10))
    assert output.shape == (5, 3, 40)


def test_lstm():
    model = torchlayers.build(torchlayers.LSTM(hidden_size=20), torch.randn(5, 3, 10))

    output, (_, _) = model(torch.randn(5, 3, 10))
    assert output.shape == (5, 3, 20)



import torch

import torchlayers as tl


def test_last_dimension_linear():
    module = tl.Linear(64)
    module = tl.build(module, torch.randn(1, 3, 32, 32))
    assert module(torch.randn(2, 6, 24, 32)).shape == (2, 6, 24, 64)



import pathlib
import tempfile

import torch

import torchlayers


def test_save():
    inputs = torch.randn(16, 32)
    temp = pathlib.Path(tempfile.gettempdir())

    layer = torchlayers.build(torchlayers.Linear(64), inputs)
    output = layer(inputs)
    torch.save(layer, temp / "linear_model.pt")

    new_layer = torch.load(temp / "linear_model.pt")
    new_output = new_layer(inputs)
    assert torch.allclose(output, new_output)


def test_convolution_save():
    inputs = torch.randn(16, 3, 32, 32)
    temp = pathlib.Path(tempfile.gettempdir())

    layer = torchlayers.build(torchlayers.Conv2d(64, kernel_size=3), inputs)
    output = layer(inputs)
    torch.save(layer, temp / "conv_model.pt")

    new_layer = torch.load(temp / "conv_model.pt")
    new_output = new_layer(inputs)
    assert torch.allclose(output, new_output)


def test_dimension_save():
    inputs = torch.randn(16, 3, 32, 32)
    temp = pathlib.Path(tempfile.gettempdir())

    layer = torchlayers.build(torchlayers.Conv(64), inputs)
    output = layer(inputs)
    torch.save(layer, temp / "conv_model.pt")

    new_layer = torch.load(temp / "conv_model.pt")
    new_output = new_layer(inputs)
    assert torch.allclose(output, new_output)



import torch

from . import module


class StochasticDepth(torch.nn.Module):
    """Randomly skip module during training with specified `p`, leaving inference untouched.

    Originally proposed by Gao Huang et al. in
    `Deep Networks with Stochastic Depth <www.arxiv.org/abs/1512.03385>`__.

    Originally devised as regularization, though `other research <https://web.stanford.edu/class/cs331b/2016/projects/kaplan_smith_jiang.pdf>`__  suggests:

    - "[...] StochasticDepth Nets are less tuned for low-level feature extraction but more tuned for higher level feature differentiation."
    - "[...] Stochasticity does not help with the ”dead neurons” problem; in fact the problem is actually more pronounced in the early layers. Nonetheless, the Stochastic Depth Network has relatively fewer dead neurons in later layers."

    It might be useful to employ this technique to layers closer to the bottleneck.

    Example::


        import torchlayers as tl

        # Assume only 128 channels can be an input in this case
        block = tl.StochasticDepth(tl.Conv(128), p=0.3)
        # May skip tl.Conv with 0.3 probability
        block(torch.randn(1, 3, 32, 32))

    Parameters
    ----------
    module: torch.nn.Module
        Any module whose output might be skipped
        (output shape of it has to be equal to the shape of inputs).
    p: float, optional
        Probability of survival (e.g. the layer will be kept). Default: ``0.5``

    """

    def __init__(self, module: torch.nn.Module, p: float = 0.5):
        super().__init__()
        if not 0 < p < 1:
            raise ValueError(
                "Stochastic Depth p has to be between 0 and 1 but got {}".format(p)
            )
        self.module: torch.nn.Module = module
        self.p: float = p
        self._sampler = torch.Tensor(1)

    def forward(self, inputs):
        if self.training and self._sampler.uniform_():
            return inputs
        return self.p * self.module(inputs)


class Dropout(module.InferDimension):
    """Randomly zero out some of the tensor elements.

    .. note::
            Changes input only if `module` is in `train` mode.

    Based on input shape it either creates `2D` or `3D` version of dropout for inputs of shape
    `4D`, `5D` respectively (including batch as first dimension).
    For every other dimension, standard `torch.nn.Dropout` will be used.

    Parameters
    ----------
    p: float, optional
        Probability of an element to be zeroed. Default: ``0.5``
    inplace: bool, optional
        If ``True``, will do this operation in-place. Default: ``False``

    """

    def __init__(self, p=0.5, inplace=False):
        super().__init__(
            dispatcher={
                5: torch.nn.Dropout3d,
                4: torch.nn.Dropout2d,
                "*": torch.nn.Dropout,
            },
            p=p,
            inplace=inplace,
        )

    # Dropout can have any input shape according to documentation
    def _module_not_found(self, inputs):
        return torch.nn.Dropout


class StandardNormalNoise(torch.nn.Module):
    """Add noise from standard normal distribution during forward pass.

    .. note::
            Changes input only if `module` is in `train` mode.

    Example::


        import torchlayers as tl

        model = tl.Sequential(
            tl.StandardNormalNoise(), tl.Linear(10), tl.ReLU(), tl.tl.Linear(1)
        )
        tl.build(model, torch.randn(3, 30))

        # Noise from Standard Normal distribution will be added
        model(torch.randn(3, 30))

        model.eval()
        # Eval mode, no noise added
        model(torch.randn(3, 30))

    """

    def forward(self, inputs):
        if self.training:
            return inputs + torch.randn_like(inputs)
        return inputs


class UniformNoise(torch.nn.Module):
    """Add noise from uniform `[0, 1)` distribution during forward pass.

    .. note::
            Changes input only if `module` is in `train` mode.

    Example::


        import torchlayers as tl

        noisy_linear_regression = tl.Sequential(
            tl.UniformNoise(), tl.Linear(1)
        )
        tl.build(model, torch.randn(1, 10))

        # Noise from Uniform distribution will be added
        model(torch.randn(64, 10))

        model.eval()
        # Eval mode, no noise added
        model(torch.randn(64, 10))

    """

    def forward(self, inputs):
        if self.training:
            return inputs + torch.rand_like(inputs)
        return inputs



import pathlib
import tempfile

import torch

import pytest
import torchlayers


@pytest.fixture
def model():
    return torchlayers.Sequential(
        torchlayers.Conv(64),
        torchlayers.BatchNorm(),
        torchlayers.ReLU(),
        torchlayers.Conv(128),
        torchlayers.BatchNorm(),
        torchlayers.ReLU(),
        torchlayers.Conv(256),
        torchlayers.GlobalMaxPool(),
        torchlayers.Linear(64),
        torchlayers.BatchNorm(),
        torchlayers.Linear(10),
    )


def test_functionality(model):
    # Initialize
    model = torchlayers.build(model, torch.randn(16, 3, 28, 28))

    optimizer = torch.optim.Adam(model.parameters())
    criterion = torch.nn.CrossEntropyLoss()


#     for _ in range(16):
#         output = model(torch.randn(16, 3, 28, 28))
#         loss = criterion(output, torch.randint(2, (16,)))
#         loss.backward()

#         optimizer.zero_grad()


def test_print_pre_init(model):
    target = r"""Sequential(
  (0): Conv(in_channels=?, out_channels=64, kernel_size=3, stride=1, padding=same, dilation=1, groups=1, bias=True, padding_mode=zeros)
  (1): BatchNorm(num_features=?, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
  (2): ReLU()
  (3): Conv(in_channels=?, out_channels=128, kernel_size=3, stride=1, padding=same, dilation=1, groups=1, bias=True, padding_mode=zeros)
  (4): BatchNorm(num_features=?, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
  (5): ReLU()
  (6): Conv(in_channels=?, out_channels=256, kernel_size=3, stride=1, padding=same, dilation=1, groups=1, bias=True, padding_mode=zeros)
  (7): GlobalMaxPool()
  (8): Linear(in_features=?, out_features=64, bias=True)
  (9): BatchNorm(num_features=?, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
  (10): Linear(in_features=?, out_features=10, bias=True)
)"""

    assert target == str(model)


def test_attribute_access_existing():
    layer = torchlayers.Conv(64)
    assert layer.kernel_size == 3
    assert layer.padding == "same"


def test_attribute_access_notinstantiated():
    layer = torchlayers.Conv(64)
    with pytest.raises(AttributeError):
        non_instantiated_channels = layer.in_channels

    layer(torch.randn(1, 8, 28, 28))
    assert layer.in_channels == 8



import inspect
import io
import types
import typing
import warnings

import torch

from . import (_dev_utils, activations, convolution, normalization, pooling,
               regularization, upsample)
from ._version import __version__
from .module import InferDimension

__all__ = ["build", "infer", "Lambda", "Reshape", "Concatenate"]


def build(module, *args, **kwargs):
    """Build PyTorch layer or module by providing example input.

    This method should be used **always** after creating module using `torchlayers`
    and shape inference especially.

    Works similarly to `build` functionality provided by `keras`.

    Provided module will be "compiled" to PyTorch primitives to remove any
    overhead.

    `torchlayers` also supports `post_build` function to perform some action after
    shape was inferred (weight initialization example below)::


        import torch
        import torchlayers as tl

        class _MyModuleImpl(torch.nn.Linear):
            def post_build(self):
                # You can do anything here really
                torch.nn.init.eye_(self.weights)

        MyModule = tl.infer(_MyModuleImpl)

    `post_build` should have no arguments other than `self` so all necessary
    data should be saved in `module` beforehand.

    Parameters
    ----------
    module : torch.nn.Module
        Instance of module to build
    *args
        Arguments required by module's `forward`
    **kwargs
        Keyword arguments required by module's `forward`
    """

    def torch_compile(module):
        with io.BytesIO() as buffer:
            torch.save(module, buffer)
            return torch.load(io.BytesIO(buffer.getvalue()))

    def run_post(module):
        for submodule in module.modules():
            function = getattr(submodule, "post_build", None)
            if function is not None:
                post_build = getattr(submodule, "post_build")
                if not callable(post_build):
                    raise ValueError(
                        "{}'s post_build is required to be a method.".format(submodule)
                    )
                submodule.post_build()

    with torch.no_grad():
        module.eval()
        module(*args, **kwargs)
    module.train()
    module = torch_compile(module)
    run_post(module)
    return module


def infer(module_class, index: str = 1):
    """Allows custom user modules to infer input shape.

    Input shape should be the first argument after `self`.

    Usually used as class decorator, e.g.::

        import torch
        import torchlayers as tl

        class _StrangeLinearImpl(torch.nn.Linear):
            def __init__(self, in_features, out_features, bias: bool = True):
                super().__init__(in_features, out_features, bias)
                self.params = torch.nn.Parameter(torch.randn(out_features))

            def forward(self, inputs):
                super().forward(inputs) + self.params

        # Now you can use shape inference of in_features
        StrangeLinear = tl.infer(_StrangeLinearImpl)

        # in_features can be inferred
        layer = StrangeLinear(out_features=64)


    Parameters
    ----------
    module_class: torch.nn.Module
        Class of module to be updated with shape inference capabilities.

    index: int, optional
        Index into `tensor.shape` input which should be inferred, e.g. tensor.shape[1].
        Default: `1` (`0` being batch dimension)

    """

    init_arguments = [
        str(argument)
        for argument in inspect.signature(module_class.__init__).parameters.values()
    ]

    # Other argument than self
    if len(init_arguments) > 1:
        name = module_class.__name__
        infered_module = type(
            name, (torch.nn.Module,), {_dev_utils.infer.MODULE_CLASS: module_class},
        )
        parsed_arguments, uninferable_arguments = _dev_utils.infer.parse_arguments(
            init_arguments, infered_module
        )

        setattr(
            infered_module, "__init__", _dev_utils.infer.create_init(parsed_arguments),
        )

        setattr(
            infered_module,
            "forward",
            _dev_utils.infer.create_forward(
                _dev_utils.infer.MODULE,
                _dev_utils.infer.MODULE_CLASS,
                parsed_arguments,
                index,
            ),
        )
        setattr(
            infered_module,
            "__repr__",
            _dev_utils.infer.create_repr(
                _dev_utils.infer.MODULE, **uninferable_arguments
            ),
        )
        setattr(
            infered_module,
            "__getattr__",
            _dev_utils.infer.create_getattr(_dev_utils.infer.MODULE),
        )

        setattr(
            infered_module,
            "__reduce__",
            _dev_utils.infer.create_reduce(_dev_utils.infer.MODULE, parsed_arguments),
        )

        return infered_module

    return module_class


class Lambda(torch.nn.Module):
    """Use any function as `torch.nn.Module`

    Simple proxy which allows you to use your own custom in
    `torch.nn.Sequential` and other requiring `torch.nn.Module` as input::

        import torch
        import torchlayers as tl

        model = torch.nn.Sequential(tl.Lambda(lambda tensor: tensor ** 2))
        model(torch.randn(64 , 20))

    Parameters
    ----------
    function : Callable
        Any user specified function

    Returns
    -------
    Any
        Anything `function` returns

    """

    def __init__(self, function: typing.Callable):
        super().__init__()
        self.function: typing.Callable = function

    def forward(self, *args, **kwargs) -> typing.Any:
        return self.function(*args, **kwargs)


class Concatenate(torch.nn.Module):
    """Concatenate list of tensors.

    Mainly useful in `torch.nn.Sequential` when previous layer returns multiple
    tensors, e.g.::

        import torch
        import torchlayers as tl

        class Foo(torch.nn.Module):
            # Return same tensor three times
            # You could explicitly return a list or tuple as well
            def forward(tensor):
                return tensor, tensor, tensor


        model = torch.nn.Sequential(Foo(), tl.Concatenate())
        model(torch.randn(64 , 20))

    All tensors must have the same shape (except in the concatenating dimension).

    Parameters
    ----------
    dim : int
        Dimension along which tensors will be concatenated

    Returns
    -------
    torch.Tensor
        Concatenated tensor along specified `dim`.

    """

    def __init__(self, dim: int):
        super().__init__()
        self.dim: int = dim

    def forward(self, inputs):
        return torch.cat(inputs, dim=self.dim)


class Reshape(torch.nn.Module):
    """Reshape tensor excluding `batch` dimension

    Reshapes input `torch.Tensor` features while preserving batch dimension.
    Standard `torch.reshape` values (e.g. `-1`) are supported, e.g.::

        import torch
        import torchlayers as tl

        layer = tl.Reshape(20, -1)
        layer(torch.randn(64, 80)) # shape (64, 20, 4)

    All tensors must have the same shape (except in the concatenating dimension).
    If possible, no copy of `tensor` will be performed.

    Parameters
    ----------
    shapes: *int
        Variable length list of shapes used in view function

    Returns
    -------
    torch.Tensor
        Concatenated tensor

    """

    def __init__(self, *shapes: int):
        super().__init__()
        self.shapes: typing.Tuple[int] = shapes

    def forward(self, tensor):
        return torch.reshape(tensor, (tensor.shape[0], *self.shapes))


###############################################################################
#
#                       MODULE AND SHAPE INFERENCE
#
###############################################################################


modules_map = {
    # PyTorch specific tensor index
    "Linear": (torch.nn, -1),
    "RNN": (torch.nn, 2),
    "LSTM": (torch.nn, 2),
    "GRU": (torch.nn, 2),
    "MultiheadAttention": (torch.nn, 2),
    "Transformer": (torch.nn, 2),
    "TransformerEncoderLayer": (torch.nn, 2),
    "TransformerDecoderLayer": (torch.nn, 2),
    # PyTorch default (1) tensor index
    "RNNCell": torch.nn,
    "LSTMCell": torch.nn,
    "GRUCell": torch.nn,
    "Conv1d": torch.nn,
    "Conv2d": torch.nn,
    "Conv3d": torch.nn,
    "ConvTranspose1d": torch.nn,
    "ConvTranspose2d": torch.nn,
    "ConvTranspose3d": torch.nn,
    "BatchNorm1d": torch.nn,
    "BatchNorm2d": torch.nn,
    "BatchNorm3d": torch.nn,
    "SyncBatchNorm": torch.nn,
    "InstanceNorm1d": torch.nn,
    "InstanceNorm2d": torch.nn,
    "InstanceNorm3d": torch.nn,
    # Torchlayers convolution
    "SqueezeExcitation": convolution,
    "Fire": convolution,
    "Conv": convolution,
    "ConvTranspose": convolution,
    "DepthwiseConv": convolution,
    "SeparableConv": convolution,
    "InvertedResidualBottleneck": convolution,
    # Torchlayers normalization
    "BatchNorm": normalization,
    "InstanceNorm": normalization,
    "GroupNorm": normalization,
    # Torchlayers upsample
    "ConvPixelShuffle": upsample,
}


def __dir__():
    return [key for key in modules_map if not key.startswith("*")] + dir(torch.nn)


def __getattr__(name: str):
    def infer_module(module, tensor_index):
        klass = getattr(module, name, None)
        if klass is not None:
            return infer(klass, tensor_index)
        return None

    def process_entry(data) -> typing.Optional:
        if isinstance(data, typing.Iterable):
            if len(data) != 2:
                raise AttributeError(
                    f"Value of to-be-inferred module: {name} second argument has "
                    f"to be of length 2 but got {len(data)}. Check torchlayers documentation."
                )
            return infer_module(*data)

        if isinstance(data, types.ModuleType):
            return infer_module(data, 1)

        raise AttributeError(
            "Torchlayers recognized entry to infer but it's entry is of incorrect type. "
            "Check documentation about registration of user modules for more info."
        )

    def registered_module():
        for key, data in modules_map.items():
            if key.startswith("*"):
                klass = process_entry(data)
                if klass is not None:
                    return klass

        return None

    def noninferable_torchlayers():
        for module in (
            activations,
            convolution,
            normalization,
            pooling,
            regularization,
            upsample,
        ):
            klass = getattr(module, name, None)
            if klass is not None:
                return klass

        return None

    data = modules_map.get(name)
    if data is None:
        klass = registered_module()
    else:
        klass = process_entry(data)

    # Try to return from torchlayers without inference
    if klass is None:
        klass = noninferable_torchlayers()

    # Try to return from torch.nn without inference
    if klass is None:
        klass = getattr(torch.nn, name, None)

    # As far as we know there is no such module
    if klass is None:
        raise AttributeError(f"module {__name__} has no attribute {name}")
    return klass



import torch

import torchlayers


def test_odd_padding():
    for kernel_size in range(1, 14, 2):
        layer = torchlayers.Conv(64, kernel_size=kernel_size)
        assert layer(torch.rand(1, 3, 28, 28)).shape == (1, 64, 28, 28)



import collections
import itertools
import math
import typing

import torch

from . import _dev_utils, module, normalization, pooling

# TO-DO: Ensure torch.jit compatibility of layers below


class _Conv(module.InferDimension):
    """Base layer for convolution-like modules.

    Passes all arguments to `_dev_utils.modules.module.InferDimension` and sets up
    `same` padding.

    See concrete classes (e.g. `Conv` or `DepthwiseConv`) for specific use cases.

    Parameters
    ----------
    dispatcher: typing.Dict[int, typing.Any]
        Dispatching dictionary for dimensionality reduction
    **kwargs
        Any arguments needed for class creation.

    """

    def __init__(
        self, dispatcher: typing.Dict[int, typing.Any], **kwargs,
    ):
        super().__init__(dispatcher=dispatcher, initializer=self._pad, **kwargs)

    def _pad(self, inner_class, inputs, **kwargs):
        def _dimension_pad(dimension, kernel_size, stride, dilation):
            if kernel_size % 2 == 0:
                raise ValueError(
                    'Only asymmetric "same" padding is currently supported. `kernel_size` size has to be odd, but got `kernel_size={}`'.format(
                        kernel_size
                    )
                )
            if stride % 2 == 0:
                raise ValueError(
                    'Only asymmetric "same" padding is currently supported. `stride` size has to be odd, but got `stride={}`'.format(
                        kernel_size
                    )
                )

            return math.ceil(
                (dimension * stride - dimension + dilation * (kernel_size - 1)) / 2
            )

        def _expand_if_needed(dimensions, argument):
            if isinstance(argument, collections.abc.Iterable):
                return argument
            return tuple(itertools.repeat(argument, len(dimensions)))

        if isinstance(kwargs["padding"], str) and kwargs["padding"].lower() == "same":
            dimensions = inputs.shape[2:]
            paddings = tuple(
                _dimension_pad(dimension, kernel_size, stride, dilation)
                for dimension, kernel_size, stride, dilation in zip(
                    dimensions,
                    *[
                        _expand_if_needed(dimensions, kwargs[name])
                        for name in ("kernel_size", "stride", "dilation")
                    ],
                )
            )
            kwargs["padding"] = paddings

        return inner_class(**kwargs)


class Conv(_Conv):
    """Standard convolution layer.

    Based on input shape it either creates 1D, 2D or 3D convolution for inputs of shape
    3D, 4D, 5D respectively (including batch as first dimension).

    Additional `same` `padding` mode was added and set as default.
    This mode preserves all dimensions excepts channels.


    `kernel_size` got a default value of `3`.

    Otherwise acts exactly like PyTorch's Convolution, see
    `documentation <https://pytorch.org/docs/stable/nn.html#convolution-layers>`__.

    Example::

        import torchlayers as tl


        class Classifier(tl.Module):
            def __init__(self, out_shape):
                super().__init__()
                self.conv1 = tl.Conv(64, 6)
                self.conv2 = tl.Conv(128)
                self.conv3 = tl.Conv(256)
                self.pooling = tl.GlobalMaxPool()
                self.dense = tl.Linear(out_shape)

            def forward(self, x):
                x = torch.relu(self.conv1(x))
                x = torch.relu(self.conv2(x))
                x = torch.relu(self.conv3(x))
                return self.dense(self.pooling(x))


    .. note::
                **IMPORTANT**: `same` currently works only for odd values of `kernel_size`,
                `dilation` and `stride`. If any of those is even you should explicitly pad
                your input asymmetrically with `torch.functional.pad` or a-like.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input image
    out_channels : int
        Number of channels produced by the convolution
    kernel_size : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Size of the convolving kernel. User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `3`
    stride : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Stride of the convolution. User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `3`
    padding : Union[str, int, Tuple[int, int], Tuple[int, int, int]], optional
        Padding added to both sides of the input. String "same" can be used with odd
        `kernel_size`, `stride` and `dilation`
        User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `same`
    dilation : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Spacing between kernel elements. String "same" can be used with odd
        `kernel_size`, `stride` and `dilation`
        User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `1`
    groups : int, optional
        Number of blocked connections from input channels to output channels. Default: 1
    bias : bool, optional
        If ``True``, adds a learnable bias to the output. Default: ``True``
    padding_mode : string, optional
        Accepted values `zeros` and `circular` Default: `zeros`

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size=3,
        stride=1,
        padding="same",
        dilation=1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        super().__init__(
            dispatcher={5: torch.nn.Conv3d, 4: torch.nn.Conv2d, 3: torch.nn.Conv1d},
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
        )


# Fix ConvTranspose and add same padding?
class ConvTranspose(_Conv):
    """Standard transposed convolution layer.

    Based on input shape it either creates 1D, 2D or 3D convolution (for inputs of shape
    3D, 4D, 5D including batch as first dimension).

    Otherwise acts exactly like PyTorch's Convolution, see
    `documentation <https://pytorch.org/docs/stable/nn.html#convolution-layers>`__.

    Default argument for `kernel_size` was added equal to `3`.

    .. note::
                **IMPORTANT**: `same` currently works only for odd values of `kernel_size`,
                `dilation` and `stride`. If any of those is even you should explicitly pad
                your input asymmetrically with `torch.functional.pad` or a-like.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input image
    out_channels : int
        Number of channels produced by the convolution
    kernel_size : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Size of the convolving kernel. User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `3`
    stride : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Stride of the convolution. User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `3`
    padding : Union[str, int, Tuple[int, int], Tuple[int, int, int]], optional
        Padding added to both sides of the input. String "same" can be used with odd
        `kernel_size`, `stride` and `dilation`
        User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `same`
    output_padding : int or tuple, optional
        Additional size added to one side of the output shape. Default: 0
    dilation : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Spacing between kernel elements. String "same" can be used with odd
        `kernel_size`, `stride` and `dilation`
        User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `1`
    groups : int, optional
        Number of blocked connections from input channels to output channels. Default: 1
    bias : bool, optional
        If ``True``, adds a learnable bias to the output. Default: ``True``
        dilation (int or tuple, optional): Spacing between kernel elements. Default: 1
    padding_mode : string, optional
        Accepted values `zeros` and `circular` Default: `zeros`

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size=3,
        stride=1,
        padding="same",
        output_padding=0,
        dilation=1,
        groups: int = 1,
        bias: bool = True,
        padding_mode="zeros",
    ):
        super().__init__(
            dispatcher={
                5: torch.nn.ConvTranspose3d,
                4: torch.nn.ConvTranspose2d,
                3: torch.nn.ConvTranspose1d,
            },
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            groups=groups,
            bias=bias,
            dilation=dilation,
            padding_mode=padding_mode,
        )


class DepthwiseConv(_Conv):
    """Depthwise convolution layer.

    Based on input shape it either creates 1D, 2D or 3D depthwise convolution
    for inputs of shape 3D, 4D, 5D respectively (including batch as first dimension).

    Additional `same` `padding` mode was added and set as default.
    This mode preserves all dimensions excepts channels.

    `kernel_size` got a default value of `3`.

    .. note::
                **IMPORTANT**: `same` currently works only for odd values of `kernel_size`,
                `dilation` and `stride`. If any of those is even you should explicitly pad
                your input asymmetrically with `torch.functional.pad` or a-like.

    .. note::
                **IMPORTANT**: `out_channels` has to be divisible by `in_channels` without remainder
                (e.g. `out_channels=64` and `in_channels=32`), otherwise error is thrown.


    Parameters
    ----------
    in_channels : int
        Number of channels in the input image
    out_channels : int
        Number of channels produced by the convolution
    kernel_size : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Size of the convolving kernel. User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `3`
    stride : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Stride of the convolution. User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `3`
    padding : Union[str, int, Tuple[int, int], Tuple[int, int, int]], optional
        Padding added to both sides of the input. String "same" can be used with odd
        `kernel_size`, `stride` and `dilation`
        User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `same`
    dilation : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Spacing between kernel elements. String "same" can be used with odd
        `kernel_size`, `stride` and `dilation`
        User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `1`
    bias : bool, optional
        If ``True``, adds a learnable bias to the output. Default: ``True``
    padding_mode : string, optional
        Accepted values `zeros` and `circular` Default: `zeros`

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size=3,
        stride=1,
        padding="same",
        dilation=1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        if out_channels % in_channels != 0:
            raise ValueError(
                "Depthwise separable convolution needs out_channels divisible by in_channels without remainder."
            )

        super().__init__(
            dispatcher={5: torch.nn.Conv3d, 4: torch.nn.Conv2d, 3: torch.nn.Conv1d},
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=in_channels,
            bias=bias,
            padding_mode=padding_mode,
        )


class SeparableConv(torch.nn.Module):
    """Separable convolution layer (a.k.a. depthwise separable convolution).

    Based on input shape it either creates 1D, 2D or 3D separable convolution
    for inputs of shape 3D, 4D, 5D respectively (including batch as first dimension).

    Additional `same` `padding` mode was added and set as default.
    This mode preserves all dimensions excepts channels.

    `kernel_size` got a default value of `3`.

    .. note::
                **IMPORTANT**: `same` currently works only for odd values of `kernel_size`,
                `dilation` and `stride`. If any of those is even you should explicitly pad
                your input asymmetrically with `torch.functional.pad` or a-like.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input image
    out_channels : int
        Number of channels produced by the convolution
    kernel_size : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Size of the convolving kernel. User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `3`
    stride : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Stride of the convolution. User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `3`
    padding : Union[str, int, Tuple[int, int], Tuple[int, int, int]], optional
        Padding added to both sides of the input. String "same" can be used with odd
        `kernel_size`, `stride` and `dilation`
        User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `same`
    dilation : Union[int, Tuple[int, int], Tuple[int, int, int]], optional
        Spacing between kernel elements. String "same" can be used with odd
        `kernel_size`, `stride` and `dilation`
        User can specify `int` or 2-tuple (for `Conv2d`)
        or 3-tuple (for `Conv3d`). Default: `1`
    bias : bool, optional
        If ``True``, adds a learnable bias to the output. Default: ``True``
    padding_mode : string, optional
        Accepted values `zeros` and `circular` Default: `zeros`

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size=3,
        stride=1,
        padding="same",
        dilation=1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        super().__init__()

        self.in_channels: int = in_channels
        self.out_channels: int = out_channels
        self.kernel_size: typing.Union[
            int, typing.Tuple[int, int], typing.Tuple[int, int, int]
        ] = kernel_size
        self.stride: typing.Union[
            int, typing.Tuple[int, int], typing.Tuple[int, int, int]
        ] = stride
        self.padding: typing.Union[
            str, int, typing.Tuple[int, int], typing.Tuple[int, int, int]
        ] = padding
        self.dilation: typing.Union[
            int, typing.Tuple[int, int], typing.Tuple[int, int, int]
        ] = dilation
        self.bias: bool = bias
        self.padding_mode: str = padding_mode

        self.depthwise = Conv(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=in_channels,
            bias=bias,
            padding_mode=padding_mode,
        )

        self.pointwise = Conv(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            dilation=1,
            groups=1,
            bias=False,
            padding_mode=padding_mode,
        )

    def forward(self, inputs):
        return self.pointwise(self.depthwise(inputs))


class ChannelShuffle(torch.nn.Module):
    """Shuffle output channels.

    When using group convolution knowledge transfer between next layers is reduced
    (as the same input channels are convolved with the same output channels).

    This layer reshuffles output channels via simple `reshape` in order to mix the representation
    from separate groups and improve knowledge transfer.

    Originally proposed by Xiangyu Zhang et al. in:
    `ShuffleNet: An Extremely Efficient Convolutional Neural Network for Mobile Devices <https://arxiv.org/abs/1707.01083>`__

    Example::


        import torchlayers as tl

        model = tl.Sequential(
            tl.Conv(64),
            tl.Swish(),
            tl.Conv(128, groups=16),
            tl.ChannelShuffle(groups=16),
            tl.Conv(256),
            tl.GlobalMaxPool(),
            tl.Linear(10),
        )

    Parameters
    ----------
    groups : int
        Number of groups used in the previous convolutional layer.

    """

    def __init__(self, groups: int):
        super().__init__()
        self.groups: int = groups

    def forward(self, inputs):
        return (
            inputs.reshape(inputs.shape[0], self.groups, -1, *inputs.shape[2:])
            .transpose(1, 2)
            .reshape(*inputs.shape)
        )


class ChannelSplit(torch.nn.Module):
    """Convenience layer splitting tensor using `p`.

    Returns two outputs, splitted accordingly to parameters.

    Example::

        import torchlayers as tl


        class Net(tl.Module):
            def __init__(self):
                super().__init__()
                self.layer = tl.Conv(256, groups=16)
                self.splitter = tl.ChannelSplit(0.5)

            def forward(x):
                outputs = self.layer(x)
                half, rest = self.splitter(outputs)
                return half # for some reason


    Parameters
    ----------
    p : float
        Percentage of channels to go into first group
    dim : int, optional
        Dimension along which input will be splitted. Default: `1` (channel dimension)

    """

    def __init__(self, p: float, dim: int = 1):
        super().__init__()
        if not 0.0 < p < 1.0:
            raise ValueError(
                "Ratio of small expand fire module has to be between 0 and 1."
            )

        self.p: float = p
        self.dim: int = dim

    def forward(self, inputs):
        return torch.split(inputs, int(inputs.shape[1] * self.p), dim=self.dim)


class Residual(torch.nn.Module):
    """Residual connection adding input to output of provided module.

    Originally proposed by He et al. in `ResNet <www.arxiv.org/abs/1512.03385>`__

    For correct usage it is advised to keep input line (skip connection) without
    any layer or activation and implement transformations only in module arguments
    (as per `Identity Mappings in Deep Residual Networks <https://arxiv.org/pdf/1603.05027.pdf>`__).

    Example::


        import torch
        import torchlayers as tl

        # ResNet-like block
        class _BlockImpl(tl.Module):
            def __init__(self, in_channels: int):
                self.block = tl.Residual(
                    tl.Sequential(
                        tl.Conv(in_channels),
                        tl.ReLU(),
                        tl.Conv(4 * in_channels),
                        tl.ReLU(),
                        tl.Conv(in_channels),
                    )
                )

            def forward(self, x):
                return self.block(x)


        Block = tl.infer(_BlockImpl)



    Parameters
    ----------
    module : torch.nn.Module
        Convolutional PyTorch module (or other compatible module).
        Shape of module's `inputs` has to be equal to it's `outputs`, both
        should be addable `torch.Tensor` instances.
    projection : torch.nn.Module, optional
        If shapes of `inputs` and `module` results are different, it's user
        responsibility to add custom `projection` module (usually `1x1` convolution).
        Default: `None`

    """

    def __init__(self, module: torch.nn.Module, projection: torch.nn.Module = None):
        super().__init__()
        self.module: torch.nn.Module = module
        self.projection: torch.nn.Module = projection

    def forward(self, inputs):
        output = self.module(inputs)
        if self.projection is not None:
            inputs = self.projections(inputs)
        return output + inputs


class Dense(torch.nn.Module):
    """Dense residual connection concatenating input channels and output channels of provided module.

    Originally proposed by Gao Huang et al. in `Densely Connected Convolutional Networks <https://arxiv.org/abs/1608.06993>`__

    Can be used just like `torchlayers.convolution.Residual` but concatenates
    channels (dimension can be specified) instead of adding.

    Parameters
    ----------
    module : torch.nn.Module
        Convolutional PyTorch module (or other compatible module).
        Shape of module's `inputs` has to be equal to it's `outputs`, both
        should be addable `torch.Tensor` instances.
    dim : int, optional
        Dimension along which `input` and module's `output` will be concatenated.
        Default: `1` (channel-wise)

    """

    def __init__(self, module: torch.nn.Module, dim: int = 1):
        super().__init__()
        self.module: torch.nn.Module = module
        self.dim: int = dim

    def forward(self, inputs):
        return torch.cat(self.module(inputs), inputs, dim=self.dim)


class Poly(torch.nn.Module):
    """Apply one module to input multiple times and sum.

    It's equation for `order` equal to :math:`N` could be expressed as

    .. math::

        I + F + F^2 + ... + F^N

    where :math:`I` is identity mapping and :math:`F` is output of `module` applied :math:`^N` times.

    Originally proposed by Xingcheng Zhang et al. in
    `PolyNet: A Pursuit of Structural Diversity in Very Deep Networks <https://arxiv.org/abs/1608.06993>`__

    Example::

        import torchlayers as tl

        # Any input will be passed 3 times
        # Through the same convolutional layer (weights and biases shared)
        layer = tl.Sequential(tl.Conv(64), tl.Poly(tl.Conv(64), order=3))
        layer(torch.randn(1, 3, 32, 32))

    Above can be rewritten by the following::

        x = torch.randn(1, 3, 32, 32)

        first_convolution = tl.Conv(64)
        output = first_convolution(x)

        shared_convolution = tl.Conv(64)
        first_level = shared_convolution(output)
        second_level = shared_convolution(first_level)
        third_level = shared_convolution(second_level)

        # That's what tl.Poly would return
        final = output + first_level + second_level + third_level


    Parameters
    ----------
    module : torch.nn.Module
        Convolutional PyTorch module (or other compatible module).
        `inputs` shape has to be equal to it's `output` shape
        (for 2D convolution it would be :math:`(C, H, W)`)
    order : int, optional
        Order of PolyInception module. For order equal to `1` acts just like
        ResNet, order of `2` was used in original paper. Default: `2`
    """

    def __init__(self, module: torch.nn.Module, order: int = 2):
        super().__init__()
        if order < 1:
            raise ValueError("Order of Poly cannot be less than 1.")

        self.module: torch.nn.Module = module
        self.order: int = order

    def extra_repr(self):
        return f"order={self.order},"

    def forward(self, inputs):
        outputs = [self.module(inputs)]
        for _ in range(1, self.order):
            outputs.append(self.module(outputs[-1]))
        return torch.stack([inputs] + outputs, dim=0).sum(dim=0)


class MPoly(torch.nn.Module):
    """Apply multiple modules to input multiple times and sum.

    It's equation for `poly_modules` length equal to :math:`N` could be expressed by

    .. math::

        I + F_1 + F_1(F_0) + ... + F_N(F_{N-1}...F_0)

    where :math:`I` is identity and consecutive :math:`F_N` are consecutive modules
    applied to output of previous ones.

    Originally proposed by Xingcheng Zhang et al. in
    `PolyNet: A Pursuit of Structural Diversity in Very Deep Networks <https://arxiv.org/abs/1608.06993>`__

    Parameters
    ----------
    *poly_modules : torch.nn.Module
        Variable arg of modules to use. If empty, acts as an identity.
        For single module acts like `ResNet`. `2` was used in original paper.
        All modules need `inputs` and `outputs` of equal `shape`.

    """

    def __init__(self, *poly_modules: torch.nn.Module):
        super().__init__()
        self.poly_modules: torch.nn.Module = torch.nn.ModuleList(poly_modules)

    def forward(self, inputs):
        outputs = [self.poly_modules[0](inputs)]
        for module in self.poly_modules[1:]:
            outputs.append(module(outputs[-1]))
        return torch.stack([inputs] + outputs, dim=0).sum(dim=0)


class WayPoly(torch.nn.Module):
    """Apply multiple modules to input and sum.

    It's equation for `poly_modules` length equal to :math:`N` could be expressed by

    .. math::

        I + F_1(I) + F_2(I) + ... + F_N

    where :math:`I` is identity and consecutive :math:`F_N` are consecutive `poly_modules`
    applied to input.

    Could be considered as an extension of standard `ResNet` to many parallel modules.

    Originally proposed by Xingcheng Zhang et al. in
    `PolyNet: A Pursuit of Structural Diversity in Very Deep Networks <https://arxiv.org/abs/1608.06993>`__

    Parameters
    ----------
    *poly_modules : torch.nn.Module
        Variable arg of modules to use. If empty, acts as an identity.
        For single module acts like `ResNet`. `2` was used in original paper.
        All modules need `inputs` and `outputs` of equal `shape`.
    """

    def __init__(self, *poly_modules: torch.nn.Module):
        super().__init__()
        self.poly_modules: torch.nn.Module = torch.nn.ModuleList(poly_modules)

    def forward(self, inputs):
        outputs = []
        for module in self.poly_modules:
            outputs.append(module(inputs))
        return torch.stack([inputs] + outputs, dim=0).sum(dim=0)


class SqueezeExcitation(torch.nn.Module):
    """Learn channel-wise excitation maps for `inputs`.

    Provided `inputs` will be squeezed into `in_channels` via average pooling,
    passed through two non-linear layers, rescaled to :math:`[0, 1]` via `sigmoid`-like function
    and multiplied with original input channel-wise.

    Originally proposed by Xingcheng Zhang et al. in
    `Squeeze-and-Excitation Networks <https://arxiv.org/abs/1709.01507>`__

    Example::


        import torchlayers as tl

        # Assume only 128 channels can be an input in this case
        block = tl.Residual(tl.Conv(128), tl.SqueezeExcitation(), tl.Conv(128))


    Parameters
    ----------
    in_channels : int
        Number of channels in the input
    hidden : int, optional
        Size of the hidden `torch.nn.Linear` layer. Usually smaller than `in_channels`
        (at least in original research paper). Default: `1/16` of `in_channels` as
        suggested by original paper.
    activation : Callable[[Tensor], Tensor], optional
        One argument callable performing activation after hidden layer.
        Default: `torch.nn.ReLU()`
    sigmoid : Callable[[Tensor], Tensor], optional
        One argument callable squashing values after excitation.
        Default: `torch.nn.Sigmoid`

    """

    def __init__(
        self, in_channels: int, hidden: int = None, activation=None, sigmoid=None,
    ):
        super().__init__()
        self.in_channels: int = in_channels
        self.hidden: int = hidden if hidden is not None else in_channels // 16
        self.activation: typing.Callable[
            [torch.Tensor], torch.Tensor
        ] = activation if activation is not None else torch.nn.ReLU()
        self.sigmoid: typing.Callable[
            [torch.Tensor], torch.Tensor
        ] = sigmoid if sigmoid is not None else torch.nn.Sigmoid()

        self._pooling = pooling.GlobalAvgPool()
        self._first = torch.nn.Linear(in_channels, self.hidden)
        self._second = torch.nn.Linear(self.hidden, in_channels)

    def forward(self, inputs):
        excitation = self.sigmoid(
            self._second(self.activation(self._first(self._pooling(inputs))))
        )

        return inputs * excitation.view(
            *excitation.shape, *([1] * (len(inputs.shape) - 2))
        )


class Fire(torch.nn.Module):
    """Squeeze and Expand number of channels efficiently operation-wise.

    First input channels will be squeezed to `hidden` channels and :math:`1 x 1` convolution.
    After that those will be expanded to `out_channels` partially done by :math:`3 x 3` convolution
    and partially by :math:`1 x 1` convolution (as specified by `p` parameter).

    Originally proposed by Forrest N. Iandola et al. in
    `SqueezeNet: AlexNet-level accuracy with 50x fewer parameters and <0.5MB model size <https://arxiv.org/abs/1602.07360>`__

    Parameters
    ----------
    in_channels : int
        Number of channels in the input
    out_channels : int
        Number of channels produced by Fire module
    hidden_channels : int, optional
        Number of hidden channels (squeeze convolution layer).
        Default: `None` (half of `in_channels`)
    p : float, optional
        Ratio of :math:`1 x 1` convolution taken from total `out_channels`.
        The more, the more :math:`1 x 1` convolution will be used during expanding.
        Default: `0.5` (half of `out_channels`)

    """

    def __init__(
        self, in_channels: int, out_channels: int, hidden_channels=None, p: float = 0.5,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        if hidden_channels is None:
            if in_channels >= 16:
                self.hidden_channels = in_channels // 2
            else:
                self.hidden_channels = 8
        else:
            self.hidden_channels = hidden_channels

        if not 0.0 < p < 1.0:
            raise ValueError("Fire's p has to be between 0 and 1, got {}".format(p))

        self.p: float = p

        self.squeeze = Conv(in_channels, self.hidden_channels, kernel_size=1)

        small_out_channels = int(out_channels * self.p)
        self.expand_small = Conv(
            self.hidden_channels, small_out_channels, kernel_size=1
        )
        self.expand_large = Conv(
            self.hidden_channels,
            out_channels - small_out_channels,
            kernel_size=3,
            padding=1,
        )

    def forward(self, inputs):
        squeeze = self.squeeze(inputs)
        return torch.cat(
            (self.expand_small(squeeze), self.expand_large(squeeze)), dim=1
        )


class InvertedResidualBottleneck(torch.nn.Module):
    """Inverted residual block used in MobileNetV2, MNasNet, Efficient Net and other architectures.

    Originally proposed by Mark Sandler et al. in
    `MobileNetV2: Inverted Residuals and Linear Bottlenecks <0.5MB model size <https://arxiv.org/abs/1801.04381>`__

    Expanded with `SqueezeExcitation` after depthwise convolution by Mingxing Tan et al. in
    `MnasNet: Platform-Aware Neural Architecture Search for Mobile <https://arxiv.org/abs/1807.11626>`__

    Due to it's customizable nature blocks from other research papers could be easily produced, e.g.
    `Searching for MobileNetV3 <https://arxiv.org/pdf/1905.02244.pdf>`__ by providing
    `torchlayers.HardSwish()` as `activation`, `torchlayers.HardSigmoid()` as `squeeze_excitation_activation`
    and `squeeze_excitation_hidden` equal to `hidden_channels // 4`.

    Parameters
    ----------
    in_channels: int
        Number of channels in the input
    hidden_channels: int, optional
        Number of hidden channels (expanded). Should be greater than `in_channels`, usually
        by factor of `4`. Default: `in_channels * 4`
    activation: typing.Callable, optional
        One argument callable performing activation after hidden layer.
        Default: `torch.nn.ReLU6()`
    batchnorm: bool, optional
        Whether to apply Batch Normalization layer after initial convolution,
        depthwise expanding part (before squeeze excitation) and final squeeze.
        Default: `True`
    squeeze_excitation: bool, optional
        Whether to use standard `SqueezeExcitation` (see `SqueezeExcitation` module)
        after depthwise convolution.
        Default: `True`
    squeeze_excitation_hidden: int, optional
        Size of the hidden `torch.nn.Linear` layer. Usually smaller than `in_channels`
        (at least in original research paper). Default: `1/16` of `in_channels` as
        suggested by original paper.
    squeeze_excitation_activation: typing.Callable, optional
        One argument callable performing activation after hidden layer.
        Default: `torch.nn.ReLU()`
    squeeze_excitation_sigmoid: typing.Callable, optional
        One argument callable squashing values after excitation.
        Default: `torch.nn.Sigmoid`

    """

    def __init__(
        self,
        in_channels: int,
        hidden_channels: int = None,
        activation=None,
        batchnorm: bool = True,
        squeeze_excitation: bool = True,
        squeeze_excitation_hidden: int = None,
        squeeze_excitation_activation=None,
        squeeze_excitation_sigmoid=None,
    ):
        def _add_batchnorm(block, channels):
            if batchnorm:
                block.append(normalization.BatchNorm(channels))
            return block

        super().__init__()

        # Argument assignments
        self.in_channels: int = in_channels
        self.hidden_channels: int = hidden_channels if hidden_channels is not None else in_channels * 4
        self.activation: typing.Callable[
            [torch.Tensor], torch.Tensor
        ] = torch.nn.ReLU6() if activation is None else activation
        self.batchnorm: bool = batchnorm
        self.squeeze_excitation: bool = squeeze_excitation
        self.squeeze_excitation_hidden: int = squeeze_excitation_hidden
        self.squeeze_excitation_activation: typing.Callable[
            [torch.Tensor], torch.Tensor
        ] = squeeze_excitation_activation
        self.squeeze_excitation_sigmoid: typing.Callable[
            [torch.Tensor], torch.Tensor
        ] = squeeze_excitation_sigmoid

        # Initial expanding block
        initial = torch.nn.Sequential(
            *_add_batchnorm(
                [
                    Conv(self.in_channels, self.hidden_channels, kernel_size=1),
                    self.activation,
                ],
                self.hidden_channels,
            )
        )

        # Depthwise block
        depthwise_modules = _add_batchnorm(
            [
                Conv(
                    self.hidden_channels,
                    self.hidden_channels,
                    kernel_size=3,
                    groups=self.hidden_channels,
                ),
                self.activation,
            ],
            self.hidden_channels,
        )

        if squeeze_excitation:
            depthwise_modules.append(
                SqueezeExcitation(
                    self.hidden_channels,
                    squeeze_excitation_hidden,
                    squeeze_excitation_activation,
                    squeeze_excitation_sigmoid,
                )
            )

        depthwise = torch.nn.Sequential(*depthwise_modules)

        # Squeeze to in channels
        squeeze = torch.nn.Sequential(
            *_add_batchnorm(
                [Conv(self.hidden_channels, self.in_channels, kernel_size=1,),],
                self.in_channels,
            )
        )

        self.block = Residual(torch.nn.Sequential(initial, depthwise, squeeze))

    def forward(self, inputs):
        return self.block(inputs)



import pathlib
import tempfile

import torch

import torchlayers


def test_basic_jit():
    inputs = torch.randn(16, 3, 32, 32)

    layer = torchlayers.build(torchlayers.Conv(64), inputs)
    output = layer(inputs)
    new_model = torch.jit.script(layer)

    new_output = new_model(inputs)
    assert torch.allclose(output, new_output)


def test_basic_jit_save():
    inputs = torch.randn(16, 3, 32, 32)
    temp = pathlib.Path(tempfile.gettempdir())

    layer = torchlayers.build(torchlayers.Conv(64), inputs)
    output = layer(inputs)
    new_model = torch.jit.script(layer)

    torch.jit.save(new_model, str(temp / "jit.pt"))

    loaded_model = torch.jit.load(str(temp / "jit.pt"))
    new_output = loaded_model(inputs)
    assert torch.allclose(output, new_output)


# Test with JIT all custom modules



import itertools

import torch

import pytest
import torchlayers


class AutoEncoder(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.encoder = torchlayers.Sequential(
            torchlayers.Conv(64, kernel_size=7),
            torchlayers.activations.Swish(),  # Direct access to module .activations
            torchlayers.InvertedResidualBottleneck(squeeze_excitation=False),
            torchlayers.AvgPool(),  # shape 64 x 128 x 128, kernel_size=2 by default
            torchlayers.HardSwish(),  # Access simply through torchlayers
            torchlayers.SeparableConv(128),  # Up number of channels to 128
            torchlayers.InvertedResidualBottleneck(),  # Default with squeeze excitation
            torch.nn.ReLU(),
            torchlayers.AvgPool(),  # shape 128 x 64 x 64, kernel_size=2 by default
            torchlayers.DepthwiseConv(256),  # DepthwiseConv easier to use
            # Pass input thrice through the same weights like in PolyNet
            torchlayers.Poly(torchlayers.InvertedResidualBottleneck(), order=3),
            torchlayers.ReLU(),  # all torch.nn can be accessed via torchlayers
            torchlayers.MaxPool(),  # shape 256 x 32 x 32
            torchlayers.Fire(out_channels=512),  # shape 512 x 32 x 32
            torchlayers.SqueezeExcitation(hidden=64),
            torchlayers.InvertedResidualBottleneck(),
            torchlayers.MaxPool(),  # shape 512 x 16 x 16
            torchlayers.InvertedResidualBottleneck(squeeze_excitation=False),
            torchlayers.Dropout(),  # Default 0.5 and Dropout2d for images
            # Randomly Switch the last two layers with 0.5 probability
            torchlayers.StochasticDepth(
                torch.nn.Sequential(
                    torchlayers.InvertedResidualBottleneck(squeeze_excitation=False),
                    torchlayers.InvertedResidualBottleneck(squeeze_excitation=False),
                ),
                p=0.5,
            ),
            torchlayers.AvgPool(),  # shape 512 x 8 x 8
        )

        # Will make this one easier and repetitive
        self.decoder = torchlayers.Sequential(
            torchlayers.Poly(torchlayers.InvertedResidualBottleneck(), order=2),
            # Has ICNR initialization by default as well
            torchlayers.ConvPixelShuffle(out_channels=512, upscale_factor=2),
            # Shape 512 x 16 x 16 after PixelShuffle
            torchlayers.Poly(torchlayers.InvertedResidualBottleneck(), order=3),
            torchlayers.ConvPixelShuffle(out_channels=256, upscale_factor=2),
            torchlayers.StandardNormalNoise(),  # add Gaussian Noise
            # Shape 256 x 32 x 32
            torchlayers.Poly(torchlayers.InvertedResidualBottleneck(), order=3),
            torchlayers.ConvPixelShuffle(out_channels=128, upscale_factor=2),
            # Shape 128 x 64 x 64
            torchlayers.Poly(torchlayers.InvertedResidualBottleneck(), order=4),
            torchlayers.ConvPixelShuffle(out_channels=64, upscale_factor=2),
            # Shape 64 x 128 x 128
            torchlayers.InvertedResidualBottleneck(),
            torchlayers.Conv(256),
            torchlayers.Swish(),
            torchlayers.BatchNorm(),
            torchlayers.ConvPixelShuffle(out_channels=32, upscale_factor=2),
            # Shape 32 x 256 x 256
            torchlayers.Conv(16),
            torchlayers.Swish(),
            torchlayers.Conv(3),
            # Shape 3 x 256 x 256
        )

    def forward(self, inputs):
        return self.decoder(self.encoder(inputs))


@pytest.fixture
def classification_model():
    return torchlayers.Sequential(
        torchlayers.Conv(64),
        torchlayers.ReLU(),
        torchlayers.MaxPool(),
        torchlayers.Residual(
            torchlayers.Sequential(
                torchlayers.Conv(64, groups=16),
                torchlayers.ReLU(),
                torchlayers.GroupNorm(num_groups=4),
                torchlayers.Conv(64, groups=16),
                torchlayers.ChannelShuffle(groups=16),
                torchlayers.ReLU(),
            )
        ),
        torchlayers.SqueezeExcitation(),
        torchlayers.Sequential(
            torchlayers.Dropout(),
            torchlayers.Conv(128),
            torchlayers.ReLU(),
            torchlayers.InstanceNorm(),
        ),
        torchlayers.Poly(
            torchlayers.WayPoly(
                torchlayers.Fire(128),
                torchlayers.Fire(128),
                torchlayers.Fire(128),
                torchlayers.Fire(128),
            ),
            order=2,
        ),
        torchlayers.AvgPool(),
        torchlayers.StochasticDepth(torchlayers.Fire(128, hidden_channels=64)),
        torchlayers.ReLU(),
        torchlayers.GlobalAvgPool(),
        torchlayers.Linear(64),
    )


@pytest.fixture
def autoencoder_model():
    return AutoEncoder()


def test_text_cnn():
    model = torch.nn.Sequential(
        torchlayers.Conv(64),  # specify ONLY out_channels
        torch.nn.ReLU(),  # use torch.nn wherever you wish
        torchlayers.BatchNorm(),  # BatchNormNd inferred from input
        torchlayers.Conv(128),  # Default kernel_size equal to 3
        torchlayers.ReLU(),
        torchlayers.Conv(256, kernel_size=11),  # "same" padding as default
        torchlayers.GlobalMaxPool(),  # Known from Keras
        torchlayers.Linear(10),  # Output for 10 classes
    )

    torchlayers.build(model, torch.randn(2, 300, 1))


def test_classification(classification_model):
    classification_model = torchlayers.build(
        classification_model, torch.randn(16, 3, 28, 28)
    )
    optimizer = torch.optim.Adam(classification_model.parameters())
    criterion = torch.nn.CrossEntropyLoss()

    for _ in range(16):
        output = classification_model(torch.randn(16, 3, 28, 28))
        loss = criterion(output, torch.randint(10, (16,)))
        loss.backward()

        optimizer.zero_grad()


def test_conv_pixel_shuffle():
    model = torchlayers.build(
        torchlayers.ConvPixelShuffle(out_channels=512, upscale_factor=2),
        torch.randn(1, 256, 128, 128),
    )


def test_residual_bottleneck():
    model = torchlayers.build(
        torchlayers.InvertedResidualBottleneck(), torch.randn(1, 128, 128, 128)
    )


def test_autoencoder(autoencoder_model):
    autoencoder_model = torchlayers.build(
        autoencoder_model, torch.randn(1, 3, 256, 256)
    )

    optimizer = torch.optim.Adam(autoencoder_model.parameters())
    criterion = torch.nn.MSELoss()

    for _ in range(16):
        inputs = torch.randn(1, 3, 256, 256)
        output = autoencoder_model(inputs)
        loss = criterion(output, inputs)
        loss.backward()

        optimizer.zero_grad()


def test_same_padding_1d():
    inputs = torch.randn(1, 5, 125)
    for kernel_size, stride, dilation in itertools.product(
        range(1, 20, 2), *[range(1, 20, 2) for _ in range(2)]
    ):
        classification_model = torchlayers.build(
            torchlayers.Conv(
                5,
                kernel_size=kernel_size,
                stride=stride,
                dilation=dilation,
                padding="same",
            ),
            inputs,
        )
        output = classification_model(inputs)
        assert output.shape == inputs.shape


def test_same_padding_2d():
    inputs = torch.randn(1, 5, 100, 100)
    for kernel_size, stride, dilation in itertools.product(
        range(1, 15, 2), *[range(1, 8, 2) for _ in range(2)]
    ):
        for kernel_size2, stride2, dilation2 in itertools.product(
            range(1, 8, 2), *[range(1, 8, 2) for _ in range(2)]
        ):
            classification_model = torchlayers.build(
                torchlayers.Conv(
                    5,
                    kernel_size=(kernel_size, kernel_size2),
                    stride=(stride, stride2),
                    dilation=(dilation, dilation2),
                    padding="same",
                ),
                inputs,
            )
            output = classification_model(inputs)
            assert output.shape == inputs.shape
